@@ -818,6 +818,7 @@ main(int argc, char **argv)
 	unsigned char *buf;
 	int ret, ch, debug_level, output_hex, bytes;
 	extern char *optarg;
+	extern int optind;
 	LogLevel ll;
 
 	__progname = ssh_get_progname(argv[0]);
@@ -853,11 +854,17 @@ main(int argc, char **argv)
 		default:
 			error("Invalid commandline option");
 			usage();
+			exit(1);
 		}
 	}
-
 	log_init(argv[0], ll, SYSLOG_FACILITY_USER, 1);
 
+	if (argc != optind) {
+		error("Unexpected commandline arguments.");
+		usage();
+		exit(1);
+	}
+
 #ifdef USE_SEED_FILES
 	prng_read_seedfile();
 #endif
